@@ -29,7 +29,9 @@ use rustc_middle::{
   },
 };
 use rustc_mir_dataflow::{fmt::DebugWithContext, graphviz, Analysis, Results};
-use rustc_span::{Pos, Span, SpanData, Symbol, SyntaxContext};
+use rustc_span::{
+  source_map::SourceMap, BytePos, Pos, Span, SpanData, Symbol, SyntaxContext,
+};
 use rustc_target::abi::VariantIdx;
 use rustc_trait_selection::infer::InferCtxtExt;
 use smallvec::SmallVec;
@@ -257,6 +259,14 @@ pub fn location_to_string(location: Location, body: &Body<'_>) -> String {
 pub struct SimplifyMir;
 impl MirPass<'tcx> for SimplifyMir {
   fn run_pass(&self, _tcx: TyCtxt<'tcx>, body: &mut Body<'tcx>) {
+    let return_blocks = body
+      .all_returns()
+      .filter_map(|loc| {
+        let bb = &body.basic_blocks()[loc.block];
+        (bb.statements.len() == 0).then(|| loc.block)
+      })
+      .collect::<HashSet<_>>();
+
     for block in body.basic_blocks_mut() {
       block.statements.retain(|stmt| {
         !matches!(
@@ -273,6 +283,11 @@ impl MirPass<'tcx> for SimplifyMir {
         TerminatorKind::FalseUnwind { real_target, .. } => TerminatorKind::Goto {
           target: real_target,
         },
+        // Ensures that control dependencies can determine the independence of differnet
+        // return paths
+        TerminatorKind::Goto { target } if return_blocks.contains(&target) => {
+          TerminatorKind::Return
+        }
         _ => continue,
       }
     }
@@ -775,6 +790,7 @@ pub trait SpanExt {
   fn merge_overlaps(spans: Vec<Span>) -> Vec<Span>;
   fn to_string(&self, tcx: TyCtxt<'_>) -> String;
   fn size(&self) -> u32;
+  fn trim_leading_whitespace(&self, source_map: &SourceMap) -> Option<Vec<Span>>;
 }
 
 impl SpanExt for Span {
@@ -887,6 +903,23 @@ impl SpanExt for Span {
   fn size(&self) -> u32 {
     self.hi().0 - self.lo().0
   }
+
+  fn trim_leading_whitespace(&self, source_map: &SourceMap) -> Option<Vec<Span>> {
+    let snippet = source_map.span_to_snippet(*self).ok()?;
+    let mut spans = Vec::new();
+    let mut start = self.lo();
+    for line in snippet.split('\n') {
+      let offset = line
+        .chars()
+        .take_while(|c| c.is_whitespace())
+        .map(|c| c.len_utf8())
+        .sum::<usize>();
+      let end = (start + BytePos(line.len() as u32)).min(self.hi());
+      spans.push(self.with_lo(start + BytePos(offset as u32)).with_hi(end));
+      start = end + BytePos(1);
+    }
+    Some(spans)
+  }
 }
 
 pub trait SpanDataExt {
